import { curry, path, prop, identity, length, assocPath, dissocPath, mergeDeepRight, compose } from 'ramda';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

var ZalgoPromise = require('zalgo-promise').ZalgoPromise;
var Promise$1 = ZalgoPromise;
var all = ZalgoPromise.all, resolve = ZalgoPromise.resolve;
var nodeKeys = function (obj) {
    return Object.keys(obj || {}).filter(function (key) { return key && key !== '_' && key !== '#'; });
};
var nowOr = curry(function (defaultValue, promise) {
    var result;
    var resolved;
    promise.then(function (res) {
        resolved = true;
        result = res;
    });
    return resolved ? result : defaultValue;
});
var now = nowOr(undefined);
var node = function (scope, soul) {
    return new ZalgoPromise(function (ok, fail) {
        var known = scope.known(soul);
        if (typeof known !== 'undefined')
            ok(known);
        scope
            .fetch(soul)
            .then(function () { return scope.known(soul); })
            .then(ok)
            .catch(fail);
    });
};
var edge = function (scope, key, parentaccess) {
    return parentaccess.then(function (data) {
        var soul = path([key, '#'], data);
        var val = prop(key, data);
        return soul ? scope.get(soul).then() : val;
    });
};
var access = function (scope, key, paccess) {
    if (paccess === void 0) { paccess = null; }
    if (!key || key === [])
        throw new Error("bad key " + key);
    var thisaccess;
    var accesses = {};
    var get = function (gKey) { return accesses[gKey] || (accesses[gKey] = access(scope, gKey, thisaccess)); };
    var then = function (fn) { return (paccess ? edge : node)(scope, key, paccess).then(fn || identity); };
    var keys = function (fn) { return then(nodeKeys).then(fn || identity); };
    var count = function (fn) { return keys(length).then(fn || identity); };
    thisaccess = { get: get, then: then, keys: keys, souls: keys, count: count };
    return thisaccess;
};
var scope = function (_a) {
    var _b = _a.graph, defaultGraph = _b === void 0 ? {} : _b, gun = _a.gun, parentScope = _a.parentScope, _c = _a.timeout, timeout = _c === void 0 ? 10000 : _c, _d = _a.cache, cache = _d === void 0 ? null : _d, getter = _a.getter, _e = _a.noGun, noGun = _e === void 0 ? false : _e, _f = _a.onlyOnce, onlyOnce = _f === void 0 ? false : _f, _g = _a.isCacheing, isCacheing = _g === void 0 ? false : _g, _h = _a.isCached, isCached = _h === void 0 ? false : _h, _j = _a.onlyCache, onlyCache = _j === void 0 ? false : _j;
    var thisScope;
    var listeners = [];
    var promises = {};
    var cachedResults = __assign({}, (cache || {}));
    var accesses = {};
    var graph = __assign({}, defaultGraph);
    var get = function (soul) { return accesses[soul] || (accesses[soul] = access(thisScope, soul)); };
    var known = function (soul) { return (parentScope ? parentScope.known(soul) : graph[soul]); };
    var on = function (fn) { return listeners.push(fn); };
    var off = function (fn) { return (listeners = listeners.filter(function (x) { return x !== fn; })); };
    var load = function (soul, data, copy) {
        if (copy === void 0) { copy = false; }
        if (data || !(soul in graph)) {
            graph[soul] = data;
            listeners.forEach(function (fn) { return fn(soul, data); });
        }
        return data;
    };
    var fetch = function (soul) {
        return parentScope
            ? parentScope.fetch(soul)
            : (promises[soul] =
                promises[soul] ||
                    new ZalgoPromise(function (ok) {
                        var readTimeout;
                        if (!gun)
                            return ok(null);
                        var receive = function (data) {
                            var args = [];
                            for (var _i = 1; _i < arguments.length; _i++) {
                                args[_i - 1] = arguments[_i];
                            }
                            clearTimeout(readTimeout);
                            if (!data)
                                console.log.apply(console, ['blank', soul, data].concat(args));
                            ok(load(soul, data));
                        };
                        readTimeout = setTimeout(function () {
                            if (!(soul in graph)) {
                                console.log('slow query', soul);
                                receive(null);
                            }
                        }, timeout);
                        if (typeof soul !== 'string')
                            throw new Error("bad SOUL " + soul);
                        if (getter)
                            getter(soul).then(receive);
                        if (!noGun) {
                            var chain = gun.get(soul);
                            if (onlyOnce) {
                                chain.once(receive);
                            }
                            else {
                                chain.on(receive);
                            }
                            if (chain.not)
                                chain.not(function () { return receive(null); });
                        }
                    }));
    };
    var getCached = function (name) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        var key = [name].concat(args).map(function (x) { return (typeof x === 'object' ? JSON.stringify(x) : "" + x); });
        return [path(key, cachedResults), key];
    };
    var cacheQuery = function (name, queryFn) {
        var args = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            args[_i - 2] = arguments[_i];
        }
        if (parentScope)
            return parentScope.cacheQuery.apply(parentScope, [name, queryFn].concat(args));
        var _a = getCached.apply(void 0, [name].concat(args)), cached = _a[0], key = _a[1];
        if (onlyCache)
            return resolve(cached);
        return queryFn.apply(void 0, [thisScope].concat(args)).then(function (result) {
            if (isCacheing) {
                cachedResults = assocPath(key, result, cachedResults);
            }
            else {
                cachedResults = dissocPath(key, cachedResults);
            }
            return result;
        });
    };
    var cachedQuery = function (name, queryFn) {
        var args = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            args[_i - 2] = arguments[_i];
        }
        if (parentScope)
            return parentScope.cachedQuery.apply(parentScope, [name, queryFn].concat(args));
        var cached = getCached.apply(void 0, [name].concat(args))[0];
        var promise = cacheQuery.apply(void 0, [name, queryFn].concat(args));
        // if (!cached && isCached) console.log("cache miss", name, args);
        if (!isCached)
            return promise;
        return cached ? resolve(nowOr(cached, promise)) : promise;
    };
    var getCache = function () { return cachedResults; };
    var getGraph = function () { return graph; };
    var getAccesses = function () { return accesses; };
    var loadCachedResults = function (newResults) {
        cachedResults = mergeDeepRight(cachedResults, newResults);
        listeners.forEach(function (fn) { return fn(); });
    };
    thisScope = {
        on: on,
        off: off,
        get: get,
        getCache: getCache,
        known: known,
        fetch: fetch,
        cacheQuery: cacheQuery,
        cachedQuery: cachedQuery,
        parentScope: parentScope,
        getGraph: getGraph,
        getAccesses: getAccesses,
        load: load,
        loadCachedResults: loadCachedResults
    };
    return thisScope;
};
var query = function (queryFn, name) {
    if (name === void 0) { name = null; }
    var cachedQuery = function (scopeObj) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        return scopeObj.cachedQuery.apply(scopeObj, [name, queryFn].concat(args));
    };
    var doCachedQuery = name ? cachedQuery : queryFn;
    var result = name ? cachedQuery : queryFn;
    result.query = queryFn;
    result.cached = doCachedQuery;
    result.now = compose(now, doCachedQuery);
    return result;
};

export { Promise$1 as Promise, all, resolve, nowOr, now, scope, query };
//# sourceMappingURL=gun-scope.es5.js.map
